#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include "rsync.h"

#define RSYNC_PORT	1812	/* Port number of rsync server */

/* Connect client to server. Returns 0 ik ok, -1 on error */
int connect_client(char *machine, int *f_in, int *f_out)
{
  int sock;
  struct sockaddr_in server;
  struct hostent *hp, *gethostbyname();
  char buf[1024];

  /* Create socket */
  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0) {
 	fprintf(FERROR,"can't open stream socket: %s\n", strerror(errno));
  	return (-1);
  }

  /* Connect socket using name specified by command line. */
  server.sin_family = AF_INET;
  hp = gethostbyname(machine);
  if (hp == 0) { fprintf(stderr, "%s: unknown host\n", machine); return (-1); }

  /* Set up the rsync port */
  bcopy(hp->h_addr, &server.sin_addr, hp->h_length);
  server.sin_port = htons(RSYNC_PORT);

  if (connect(sock, (struct sockaddr *) & server, sizeof(server)) < 0) {
    fprintf(FERROR,"can't connect stream socket: %s\n", strerror(errno));
    return (-1);
  }

  /* File descriptor for both input and output is the socket */
  *f_in = sock; *f_out = sock;
  return (0);
}

/* fdgets is a non-stdio version of fgets, with the exception that
 * it removes \n and \r from the end of the line
 */
char *fdgets(char *str, int size, int fd)
{
  int i;
  char *s = str;

  /* Give up if str doesn't exist or is too small */
  if (str == NULL || size < 2) return (NULL);

  /* Read characters until end of buffer or we hit a \n */
  for (i = 0; i < size; i++, s++) {
    if (read(fd, s, 1) == -1) return (NULL);
    if (*s == '\n') break;
  }

  /* Terminate the string by replacing the \n, if possible */
  if (*s == '\n' && i < size) *s = '\0';

  /* Remove any \r as well */
  s--; if (i>0 && *s=='\r') *s='\0';

  /* Finally, null the last char in the buffer, for good measure */
  str[size - 1] = '\0'; return (str);
}

/* This is used instead of printf, to ensure the output is sent! */
void printfd(char *s)
{ write(1, s, strlen(s)); }


/* Lookup the named tree, check it exists, check we can chdir into it */
int chdir_tree(char *dir)
{
  FILE *treef;
  char *buf = (char *) malloc(256);
  char *treename, *realdir;

  /* Return is no tree name */
  if (dir == NULL || *dir == '\0') return (-1);

  /* Open the list of trees, return error if we can't */
  treef = fopen(RSYNCD_TREES, "r");

  if (treef == NULL) {
    syslog(LOG_NOTICE, "Could not open list of rsyncd trees");
    printfd("500 Could not open list of rsyncd trees\n");
    return (-1);
  }

  /* Get a line from the file, skipping lines starting with # */
  while (fgets(buf, 255, treef)) {
    if (buf[0] == '#') continue;

    /* Find the treename and the real directory name */
    /* Skip line if it's badly formed */
    buf[strlen(buf) - 1] = '\0';
    realdir = buf;
    treename= strsep(&realdir, " \t");
    if (realdir==NULL) continue;
    while (*realdir==' ' || *realdir=='\t') realdir++;
    if (*realdir=='\0') continue;

    /* Skip if the tree isn't the one requested */
    if (strcmp(dir, treename)) continue;

    /* Check we can chdir into it */
    if (chdir(realdir) == 0) {
        syslog(LOG_NOTICE, "Tree exists & chdir worked");
	printfd("200 Looks good, let's rsync\n");
	fclose(treef);
	return (0);
    } else {
	syslog(LOG_NOTICE, "Could not chdir to the named tree");
	printfd("501 Could not chdir to the named tree\n");
	fclose(treef);
	return (-1);
      }
  }

  /* Hmm, the request isn't in the list */
  syslog(LOG_NOTICE, "Named tree does not exist");
  printfd("502 Named tree does not exist on this server\n");
  fclose(treef);
  return (-1);
}
